var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { debounce, generateFieldSaveKey, mountElement, unmountElement } from './utils.js';
export class Field {
    constructor(parent, form, options) {
        this.options = {
            id: '',
            type: 'text',
            required: false,
            validation: (value, data, required) => {
                if (required && !value)
                    return 'This field is required';
                return true;
            },
            default: "",
            debounce: 200,
            className: 'form-input',
        };
        this.inputElement = null;
        this.containerElement = null;
        this.labelElement = null;
        this.validationElement = null;
        this.hasScheme = false;
        this._isValid = null;
        this._isDisabled = false;
        this._isRequired = false;
        this._isMounted = false;
        this._isVisible = true;
        this._vMessage = null;
        this._value = null;
        this.initializeOptions(options);
        this._id = this.options.id;
        this._form = form;
        this._parent = parent;
        this._type = options.type;
        this._saveKey = generateFieldSaveKey(this._form.getId(), this._id);
    }
    initializeOptions(options) {
        this.options = Object.assign({}, this.options, options);
    }
    initialize() {
        return __awaiter(this, void 0, void 0, function* () {
            this.load();
            this.update();
            this.bindChange();
        });
    }
    setValue(value, save = true) {
        this._value = value;
        this.syncValue();
        this._form.setData(this._id, value);
        if (save)
            this.save();
    }
    syncValue() {
        if (this.inputElement &&
            (this.inputElement instanceof HTMLInputElement || this.inputElement instanceof HTMLTextAreaElement) &&
            this.inputElement.value !== this._value) {
            this.inputElement.value = this._value;
        }
    }
    getValue() {
        return this._value;
    }
    getId() {
        return this._id;
    }
    getSaveKey() {
        return this._saveKey;
    }
    getType() {
        return this._type;
    }
    getVisibility() {
        return this._isVisible;
    }
    getValidity() {
        return this._isValid;
    }
    isDisabled() {
        return this._isDisabled;
    }
    getValidationMessage() {
        return this._vMessage;
    }
    getForm() {
        return this._form;
    }
    createContainerElement() {
        this.containerElement = document.createElement('div');
        this.containerElement.className = 'form-field ' + this._type;
        this.containerElement.setAttribute('id', this._id + '_container');
    }
    createInputElement() {
        this.inputElement = document.createElement('input');
        this.inputElement.setAttribute('id', this._id);
        this.inputElement.setAttribute('name', this.options.name || this._id);
        this.inputElement.setAttribute('type', this._type);
        if (this.options.placeholder)
            this.inputElement.setAttribute('placeholder', this.options.placeholder);
        this.inputElement.className = this.options.className;
    }
    createLabelElement() {
        this.labelElement = document.createElement('label');
        const label = document.createElement('p');
        if (this.options.label)
            label.innerText = this.options.label;
        label.setAttribute('id', this._id + '_label');
        this.labelElement.setAttribute('for', this._id);
        label.className = 'form-field-label';
        if (this.options.label)
            mountElement(label, this.labelElement);
    }
    createValidationElement() {
        this.validationElement = document.createElement('p');
        this.validationElement.setAttribute('id', this._id + '_validation');
        this.validationElement.className = 'form-field-validation';
        this.validationElement.style.display = 'none';
    }
    onGui() {
        this.createContainerElement();
        this.createInputElement();
        this.createLabelElement();
        this.createValidationElement();
        if (this.inputElement && this.labelElement)
            mountElement(this.inputElement, this.labelElement);
        if (this.labelElement && this.validationElement)
            mountElement(this.validationElement, this.labelElement);
        if (this.containerElement && this.labelElement)
            mountElement(this.labelElement, this.containerElement);
    }
    mount() {
        if (this.containerElement)
            mountElement(this.containerElement, this._parent);
        this._isMounted = true;
    }
    unmount() {
        if (this.containerElement)
            unmountElement(this.containerElement);
        this._isMounted = false;
    }
    destroy() {
        if (this._parent)
            unmountElement(this._parent);
    }
    handleValidatedField() {
        var _a, _b;
        if (!this._isValid) {
            (_a = this.labelElement) === null || _a === void 0 ? void 0 : _a.classList.add('form-error');
            if (this.validationElement && this._vMessage) {
                this.validationElement.innerText = this._vMessage;
                this.validationElement.style.display = 'block';
            }
        }
        else {
            if (this.validationElement) {
                this.validationElement.style.display = 'none';
            }
            (_b = this.labelElement) === null || _b === void 0 ? void 0 : _b.classList.remove('form-error');
        }
        this._form.updateError(this._id, this._isValid);
    }
    handleVisibility() {
        if (this._isVisible && !this._isMounted)
            this.mount();
        if (!this._isVisible && this._isMounted)
            this.unmount();
    }
    handleDisabled() {
        var _a, _b;
        if (this._isDisabled)
            (_a = this.inputElement) === null || _a === void 0 ? void 0 : _a.setAttribute('disabled', 'true');
        else
            (_b = this.inputElement) === null || _b === void 0 ? void 0 : _b.removeAttribute('disabled');
    }
    handleRequired() {
        var _a, _b;
        if (this._isRequired) {
            (_a = this.containerElement) === null || _a === void 0 ? void 0 : _a.classList.add('field-required');
        }
        else {
            (_b = this.containerElement) === null || _b === void 0 ? void 0 : _b.classList.remove('field-required');
        }
    }
    bindChange() {
        if (this.inputElement)
            this.inputElement.addEventListener('input', debounce(this.change, this.options.debounce, this));
    }
    save() {
        if (this._form.savesProgress() && this._form.hasValidLicense() && this._value !== undefined) {
            localStorage.setItem(this._saveKey, JSON.stringify(this._value));
        }
    }
    load() {
        var _a;
        if (this._form.savesProgress() && this._form.hasValidLicense()) {
            const value = localStorage.getItem(this._saveKey);
            if (value !== null) {
                this.setValue(JSON.parse(value), false);
                return;
            }
        }
        this.setValue((_a = this.options.default) !== null && _a !== void 0 ? _a : null, false);
    }
    reset() {
        return __awaiter(this, void 0, void 0, function* () {
            localStorage.removeItem(this._saveKey);
            if (this.options.default)
                this.setValue(this.options.default, false);
            else
                this.setValue(null, false);
            this.update();
        });
    }
    update() {
        return __awaiter(this, void 0, void 0, function* () {
            this.updateVisibilityBasedOnConditions();
            this.updateDisabledStatus();
            this.updateRequiredStatus();
            this.handleVisibility();
            this.handleDisabled();
            this.handleRequired();
        });
    }
    updateVisibilityBasedOnConditions() {
        if (this.options.conditions)
            this._isVisible = this.options.conditions(this._value, this._form.getData());
    }
    updateDisabledStatus() {
        if (this.options.disabled) {
            this._isDisabled =
                typeof this.options.disabled === 'function'
                    ? this.options.disabled(this._value, this._form.getData())
                    : this.options.disabled;
        }
    }
    updateRequiredStatus() {
        if (this.options.required) {
            this._isRequired =
                typeof this.options.required === 'function'
                    ? this.options.required(this._value, this._form.getData())
                    : this.options.required;
        }
    }
    validate() {
        if (!this._isVisible)
            return true;
        if (this.options.validation) {
            const validation = this.options.validation(this._value, this._form.getData(), this._isRequired);
            this._isValid = validation === true;
            this._vMessage = validation === true ? '' : validation;
            this.handleValidatedField();
        }
        else
            this._isValid = true;
        return this._isValid;
    }
    change(event) {
        this.setValue(event.target.value);
        this.validate();
        if (this.options.change)
            this.options.change(this._value);
    }
}
